#include "quantificationMethod.h"
#include "../mcq_error.h"
#include "../share/utilities.h"

QuantificationMethod::QuantificationMethod(const QString &xml_id)
    : _xml_id(xml_id) {
  if (_precision_unit == pappso::PrecisionUnit::ppm) {
    _lower_precision_p = pappso::Precision::getPpmInstance(_lower_precision);
    _upper_precision_p = pappso::Precision::getPpmInstance(_upper_precision);
  } else {
    _lower_precision_p = pappso::Precision::getDaltonInstance(_lower_precision);
    _upper_precision_p = pappso::Precision::getDaltonInstance(_upper_precision);
  }
}

QuantificationMethod::QuantificationMethod(const QuantificationMethod &other)
    : _xml_id(other._xml_id) {

  _p_xic_filter = other._p_xic_filter;
  _p_detection_method = other._p_detection_method;
  _m_xic_extract_method = other._m_xic_extract_method;
  _precision_unit = other._precision_unit;
  _upper_precision = other._upper_precision;
  _lower_precision = other._lower_precision;

  _lower_precision_p = other._lower_precision_p;
  _upper_precision_p = other._upper_precision_p;
}

QuantificationMethod::~QuantificationMethod() {
  qDebug() << "QuantificationMethod::~QuantificationMethod 1 ";
  delete _p_detection_method;
  qDebug() << "QuantificationMethod::~QuantificationMethod 2 ";
  // delete xic filters
  if (_p_xic_filter != nullptr) {
    _p_xic_filter->deleteSink();
    delete _p_xic_filter;
  }

  qDebug() << "QuantificationMethod::~QuantificationMethod end ";
}

const QString &QuantificationMethod::getXmlId() const { return _xml_id; }

void QuantificationMethod::printInfos(QTextStream &out) const {
  out << "\t Quantification method '" << _xml_id << "' parameters : " << endl;
  out << "\t xic type : '" << Utilities::toString(_m_xic_extract_method) << "'"
      << endl;
  /*
  std::vector<const FilterBase *>::const_iterator it_f;
  for (it_f = _v_p_xic_filters.begin();
          it_f != _v_p_xic_filters.end();
          ++it_f) {
      out << ( (*it_f)->printInfos() );
  }
  */
  // out << (_p_detection_method->printInfos());
}

void QuantificationMethod::setDetectionMethod(
    pappso::XicDetectionInterface *detection_method) {
  _p_detection_method = detection_method;
}

const pappso::XicDetectionInterface *
QuantificationMethod::getDetectionMethod() const {
  return _p_detection_method;
}

void QuantificationMethod::setXicExtractMethod(
    pappso::XicExtractMethod xic_extraction_method) {
  _m_xic_extract_method = xic_extraction_method;
}

pappso::XicExtractMethod QuantificationMethod::getXicExtractMethod() const {
  return _m_xic_extract_method;
}

void QuantificationMethod::addFilter(pappso::XicFilterInterface *p_filter) {
  if (_p_xic_filter == nullptr) {
    _p_xic_filter = p_filter;
  } else {
    _p_xic_filter->getLastFilter()->setSink(p_filter);
  }
}

const pappso::XicFilterInterface *QuantificationMethod::getXicFilters() const {
  return _p_xic_filter;
}
void QuantificationMethod::setXicExtractPrecision(
    pappso::PrecisionP precision) {
  _lower_precision_p = precision;
  _upper_precision_p = precision;
  _precision_unit = precision->getUnit();

  _lower_precision = precision->getValue();
  _upper_precision = precision->getValue();
}

void QuantificationMethod::setXicExtractLowerPrecision(
    pappso::pappso_double min) {
  _lower_precision = min;

  if (_precision_unit == pappso::PrecisionUnit::ppm) {
    _lower_precision_p = pappso::Precision::getPpmInstance(_lower_precision);
  } else {
    _lower_precision_p = pappso::Precision::getDaltonInstance(_lower_precision);
  }
}
void QuantificationMethod::setXicExtractUpperPrecision(
    pappso::pappso_double max) {
  _upper_precision = max;

  if (_precision_unit == pappso::PrecisionUnit::ppm) {
    _upper_precision_p = pappso::Precision::getPpmInstance(_upper_precision);
  } else {
    _upper_precision_p = pappso::Precision::getDaltonInstance(_upper_precision);
  }
}
pappso::PrecisionP QuantificationMethod::getLowerPrecision() const {
  return _lower_precision_p;
}
pappso::PrecisionP QuantificationMethod::getUpperPrecision() const {
  return _upper_precision_p;
}

pappso::PrecisionP QuantificationMethod::getMeanPrecision() const {

  if (_precision_unit == pappso::PrecisionUnit::ppm) {
    return pappso::Precision::getPpmInstance(
        (_upper_precision + _lower_precision) / 2);
  }
  return pappso::Precision::getDaltonInstance(
      (_upper_precision + _lower_precision) / 2);
}

const pappso::MassRange
QuantificationMethod::getMassRange(pappso::mz mz) const {
  return pappso::MassRange(mz, _lower_precision_p, _upper_precision_p);
}
